"""configuration/__init__.py

Created by Thomas Mangin on 2015-06-19.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# flake8: noqa: F401,E261

from __future__ import annotations

from exabgp.configuration.configuration import Configuration
